mod grpc;
use tokio_stream::{self as stream};
use tonic::{metadata::MetadataValue, transport::Channel, Request};

use grpc::dummy::{test_service_client::TestServiceClient, Empty, SomeRequest, StreamRequest};

#[tokio::main]
async fn main() -> Result<(), Box<dyn std::error::Error>> {
    let channel = Channel::from_static("http://[::1]:50051").connect().await?;
    let token: MetadataValue<_> = "Bearer some-secret-token".parse()?;
    let mut client = TestServiceClient::with_interceptor(channel, move |mut req: Request<_>| {
        println!("Request: {:?}", req);
        req.metadata_mut().insert("authorization", token.clone());
        Ok(req)
    });

    let my_real_request = SomeRequest { what_client_sends: String::from("Hello from client!") };
    let mut grpc_wrapped_request = Request::new(my_real_request.clone());
    let signature: MetadataValue<_> = format!("I am a signature of {my_real_request:?}").parse()?;
    grpc_wrapped_request.metadata_mut().insert("request-signature", signature.clone());
    let response = client.get_something(grpc_wrapped_request).await?;
    println!("Unary response: {:?}", response.get_ref().what_server_sends);

    let messages = vec![
        StreamRequest { what_client_sends: "Client stream #1".to_string() },
        StreamRequest { what_client_sends: "Client stream #2".to_string() },
        StreamRequest { what_client_sends: "Client stream #3".to_string() },
    ];

    let outbound = stream::iter(messages);

    let response = client.client_messages(Request::new(outbound)).await?;
    println!("Client streaming response: {:?}", response.get_ref());

    let mut stream = client.server_messages(Request::new(Empty {})).await?.into_inner();

    println!("Server streaming messages:");
    while let Some(item) = stream.message().await? {
        println!("  -> {}", item.what_server_sends);
    }

    println!("Server stream ended.");

    Ok(())
}
